import axios from "axios";
import { splitSignature } from "ethers/lib/utils";
import { BackendApi } from "./backend/backendApi";
import { isSmartWallet } from "./helpers/ethers";
import { isValidUrl } from "./helpers/validityHelper";
import { InvalidUrlError } from "./errors";

export class ClientLibrary {
    private _backendApi: BackendApi;

    /**
     * Construct the Client Library instance. Optionally
     * pass the backend URL to initialize directly.
     * @param newBackendUrl
     */
    constructor(newBackendUrl?: string) {
        if (newBackendUrl) {
            this.initialize(newBackendUrl);
        }
    }

    /**
     * Initialize the web3Provider by providing the URL
     * to the backend.
     * @param newBackendUrl
     * @throws InvalidUrlError
     */
    public initialize(newBackendUrl: string) {
        if (isValidUrl(newBackendUrl)) {
            this._backendApi = new BackendApi(axios.create({
                baseURL: newBackendUrl
            }));
        } else {
            throw new InvalidUrlError(newBackendUrl);
        }
    }

    /**
     * Authenticate the user by asking them to sign
     * a message. The web3Provider argument is used
     * to sign the authentication message.
     * This returns a JWT token which should be used
     * for subsequent calls.
     * @param account
     * @param web3Provider
     * @returns JWT Token
     */
    public async authenticateUser(account: string, web3Provider: any): Promise<string> {
        const nonce = await this._backendApi.generateNonce(account);

        const EIP712Domain = ClientLibrary.getEIP712Domain();
        const domain = ClientLibrary.getDomain(web3Provider.network.chainId);

        const AuthSignature: Array<{ name: string; type: string; }> = [{ name: "value", type: "string" }];

        const data: string = JSON.stringify({
            types: {
                EIP712Domain,
                AuthSignature,
            },
            domain,
            primaryType: "AuthSignature",
            message: {
                value: `Authentication message: ${nonce}`
            },
        });

        const isSmartWalletAccount = isSmartWallet(web3Provider);
        const signatureLike = await web3Provider.send("eth_signTypedData_v4", [account, data]);
        const signature = await splitSignature(signatureLike);

        return await this._backendApi.verifySignature(
            account,
            isSmartWalletAccount,
            domain,
            {AuthSignature},
            signature
        );
    }

    /**
     * This handles the sequence of interactions to
     * complete the Commit action (with backend and
     * with contracts via the contracts SDK library).
     * @param web3Provider
     * @param voucherSetId
     * @param jwt
     * @returns Success status of commit
     */
    public commitToBuy(web3Provider: any, voucherSetId: string, jwt: any): string {return ""}

    /**
     * Compose the domain object.
     * @param chainId
     * @private
     * @returns Domain object
     */
    private static getDomain(chainId: number): { name: string; version: string; chainId: any; verifyingContract: string; } {
        return {
            name: "Boson Protocol",
            version: "1",
            chainId: chainId,
            verifyingContract: "0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
        }
    }

    /**
     * Composes the domain definition object.
     * @private
     * @returns Domain definition object
     */
    private static getEIP712Domain() {
        return [
            { name: "name", type: "string" },
            { name: "version", type: "string" },
            { name: "chainId", type: "uint256" },
            { name: "verifyingContract", type: "address" },
        ];
    }
}